(*
Copyright 2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory KBO
imports
  Term_Order
  Term_Multiset
  Lexicographic_Extension
  Simple_Termination
begin

(* scf_list expands the subterm coefficient entries, so that each argument is replicates w.r.t. its scf *)
definition scf_list :: "(nat \<Rightarrow> nat) \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "scf_list scf xs = concat (map (\<lambda> (x,i). replicate (scf i) x) (zip xs [0 ..< length xs]))"

lemma scf_list_empty[simp]: "scf_list scf [] = []" unfolding scf_list_def by auto

lemma scf_list_bef_i_aft[simp]: "scf_list scf (bef @ i # aft) = 
  scf_list scf bef @ replicate (scf (length bef)) i @ scf_list (\<lambda> i. scf (Suc (length bef + i))) aft"
  unfolding scf_list_def
proof (induct aft rule: List.rev_induct)
  case Nil show ?case by simp
next
  case (snoc a aft)
  def bia \<equiv> "bef @ i # aft"
  have bia: "bef @ i # aft @ [a] = bia @ [a]" unfolding bia_def by simp
  have zip: "zip (bia @ [a]) [0..<length (bia @ [a])]
   = zip bia [0..<length bia] @ [(a,length bia)]" by simp
  have concat: "concat (map (\<lambda>(x, i). replicate (scf i) x) (zip bia [0..<length bia] @ [(a, length bia)]))
    = concat (map (\<lambda>(x, i). replicate (scf i) x) (zip bia [0..<length bia])) @ replicate (scf (length bia)) a"
    by simp
  show ?case unfolding bia zip concat
    unfolding bia_def snoc
    by simp
qed

lemma scf_list_map[simp]: "scf_list scf (map f xs) = map f (scf_list scf xs)"
  unfolding scf_list_def
  by (induct xs rule: List.rev_induct, auto)

(* scf_term replicates each argument accordings to its scf *)
fun scf_term :: "('f \<times> nat \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term" where
  "scf_term scf (Var x) = (Var x)"
| "scf_term scf (Fun f ts) = Fun f (scf_list (scf (f,length ts)) (map (scf_term scf) ts))"

lemma scf_list_subset: "set (scf_list scf xs) \<subseteq> set xs"
  by (auto simp: scf_list_def set_zip)

lemma vars_term_scf_subset: "vars_term (scf_term scf s) \<subseteq> vars_term s"
proof (induct s)
  case (Fun f ss)
  have "vars_term (scf_term scf (Fun f ss)) = (\<Union>x\<in>set (scf_list (scf (f, length ss)) ss). vars_term (scf_term scf x))" by auto
  also have "\<dots> \<subseteq> (\<Union>x\<in>set ss. vars_term (scf_term scf x))" using scf_list_subset[of _ ss] by blast
  also have "\<dots> \<subseteq> (\<Union>x\<in>set ss. vars_term x)" using Fun by auto
  finally show ?case by auto
qed auto

lemma scf_term_subst: "scf_term scf (t \<cdot> \<sigma>) = scf_term scf t \<cdot> (\<lambda> x. scf_term scf (\<sigma> x))"
proof (induct t)
  case (Fun f ts)
  {
    fix t
    assume "t \<in> set (scf_list (scf (f, length ts)) ts)"
    with scf_list_subset[of _ ts] have "t \<in> set ts" by auto
    hence "scf_term scf (t \<cdot> \<sigma>) = scf_term scf t \<cdot> (\<lambda>x. scf_term scf (\<sigma> x))"  by (rule Fun)
  }
  thus ?case by auto
qed auto

locale weight_fun =
  fixes w :: "'f \<times> nat \<Rightarrow> nat"
    and w0 :: nat
    and least :: "'f \<Rightarrow> bool"
    and scf :: "'f \<times> nat \<Rightarrow> nat \<Rightarrow> nat" 
begin

(* for the weight function we first use the one where weights are multiplied by their scf *)
fun weight :: "('f, 'v) term \<Rightarrow> nat" where
  "weight (Var x) = w0" |
  "weight (Fun f ts) =
   (let n = length ts; scff = scf (f,n) in
    w (f, n) + listsum (map (\<lambda> (ti, i). weight ti * scff i) (zip ts [0 ..< n])))"

(* afterwards, we do not multiply, but instead replicate the arguments via scf_list,
   advantage: then both weight and vars_term (scf_term .) work via scf_list *)
lemma weight_simp [simp]:
  "weight (Fun f ts) = w (f, length ts) + listsum (map weight (scf_list (scf (f, length ts)) ts))"
proof -
  {
    fix scf
    have "(\<Sum>(ti, i)\<leftarrow>zip ts [0..<length ts]. weight ti * scf i) = listsum (map weight (scf_list scf ts))"
    proof (induct ts rule: List.rev_induct)
      case Nil show ?case by auto
    next
      case (snoc t ts)
      {
        fix n
        have "listsum (replicate n (weight t)) = n * weight t"
          by (induct n, auto)
      } note repl = this
      from snoc show ?case
        by (simp add: scf_list_def repl)
    qed
  }
  thus ?thesis  
    unfolding weight.simps Let_def by auto
qed

declare weight.simps(2)[simp del]

abbreviation SCF where "SCF \<equiv> scf_term scf"

lemma listsum_scf_list:
  assumes "\<And> i. i < length ts \<Longrightarrow> f i > 0"
  shows "listsum (map weight ts) \<le> listsum (map weight (scf_list f ts))"
  using assms unfolding scf_list_def
proof (induct ts rule: List.rev_induct)
  case (snoc t ts)
  have IH: "listsum (map weight ts) \<le> listsum (map weight (concat (map (\<lambda>(x, i). replicate (f i) x) (zip ts [0..<length ts]))))"
    by (rule snoc(1), rule snoc(2), auto)
  from snoc(2)[of "length ts"] obtain n where "f (length ts) = Suc n" by (auto elim: lessE)
  with IH show ?case by simp
qed simp

(* in this version of KBO, for the Var-Var case we require identical variables,
   later in the implementation in KBO_Impl, we drop this condition as it is ensured
   by the multiset comparison of variables. *)
(* the result are two booleans: (strict decrease, non-strict decrease) *)
fun
  kbo_prec :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<Rightarrow>
    ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "kbo_prec S W s t = (if (vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> weight t \<le> weight s)
    then if (weight t < weight s)
      then (True, True)
      else (case s of
        Var y \<Rightarrow> (False, (case t of Var x \<Rightarrow> x = y | Fun g ts \<Rightarrow> ts = [] \<and> least g)) 
      | Fun f ss \<Rightarrow> (case t of
          Var x \<Rightarrow> (True, True)
        | Fun g ts \<Rightarrow> if S (f, length ss) (g, length ts)
          then (True, True)
          else if W (f, length ss) (g, length ts)
          then lex_ext_unbounded (kbo_prec S W) ss ts
          else (False, False)))
    else (False, False))"

abbreviation S_prec where "S_prec S W \<equiv> \<lambda> s t. fst (kbo_prec S W s t)"
abbreviation NS_prec where "NS_prec S W \<equiv> \<lambda> s t. snd (kbo_prec S W s t)"

lemma kbo_prec_mono:
  assumes S_imp: "\<And>fn gm. S fn gm \<Longrightarrow> S' fn gm"
    and W_imp: "\<And>fn gm. W fn gm \<Longrightarrow> W' fn gm"
  shows "(S_prec S W s t \<longrightarrow> S_prec S' W' s t) \<and> (NS_prec S W s t \<longrightarrow> NS_prec S' W' s t)"
    (is "(?S s t \<longrightarrow> ?S' s t) \<and> (?N s t \<longrightarrow> ?N' s t)")
proof (induct S\<equiv>S W\<equiv>W s t rule: kbo_prec.induct)
  case (1 s t)
  show ?case
  proof (cases "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> weight t \<le> weight s")
    case True
    note vars_weight = this
    show ?thesis
    proof (cases "weight t < weight s")
      case True with vars_weight show ?thesis by simp
    next
      case False
      note non_strict_weight = this
      show ?thesis
      proof (cases s)
        case (Var x) with vars_weight and non_strict_weight show ?thesis by auto
      next
        case (Fun f ss)
        note s = this
        show ?thesis
        proof (cases t)
          case (Var y) with vars_weight and non_strict_weight and s show ?thesis by auto
        next
          case (Fun g ts)
          note t = this
          show ?thesis
          proof (cases "S (f, length ss) (g, length ts)")
            case True
            note S = this
            with S_imp have "S' (f, length ss) (g, length ts)" by blast
            with vars_weight and S and non_strict_weight and s and t show ?thesis by auto
          next
            case False
            note not_S = this
            show ?thesis
            proof (cases "W (f, length ss) (g, length ts)")
              case False
              note not_W = this
              with vars_weight and non_strict_weight and s and t and not_S
                show ?thesis by auto
            next
              case True
              note W = this
              with W_imp have W': "W' (f, length ss) (g, length ts)" by blast
              let ?kbo = "kbo_prec S W"
              let ?kbo' = "kbo_prec S' W'"
              let ?lex = "lex_ext_unbounded ?kbo"
              let ?lex' = "lex_ext_unbounded ?kbo'"
              note 1(1) [OF vars_weight non_strict_weight s t not_S W]
              with lex_ext_unbounded_mono [of ss ts ?kbo ?kbo']
                have IH: "fst (?lex ss ts) \<Longrightarrow> fst (?lex' ss ts)"
                         "snd (?lex ss ts) \<Longrightarrow> snd (?lex' ss ts)"
                by auto
              show ?thesis
              proof (cases "S' (f, length ss) (g, length ts)")
                case True
                note S' = this
                with vars_weight and non_strict_weight and s and t
                  show ?thesis by simp
              next
                case False
                note not_S' = this
                with vars_weight and non_strict_weight and s and t and not_S and W and W'
                  show ?thesis
                  using IH by simp
              qed
            qed
          qed
        qed
      qed
    qed
  next
    case False
    then show ?thesis by auto
  qed
qed

end

locale weight_fun_prc = weight_fun w w0 least scf
  for w w0 and least :: "'f \<Rightarrow> bool" and scf +
  fixes pr_strict :: "'f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool"
    and pr_weak :: "'f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool"
begin

fun kbo :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool" where
  "kbo s t = (if (vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> weight t \<le> weight s)
    then if (weight t < weight s)
      then (True, True)
      else (case s of
        Var y \<Rightarrow> (False, (case t of Var x \<Rightarrow> x = y | Fun g ts \<Rightarrow> ts = [] \<and> least g)) 
      | Fun f ss \<Rightarrow> (case t of
          Var x \<Rightarrow> (True, True)
        | Fun g ts \<Rightarrow> if pr_strict (f, length ss) (g, length ts)
          then (True, True)
          else if pr_weak (f, length ss) (g, length ts)
          then lex_ext_unbounded kbo ss ts
          else (False, False)))
    else (False, False))"

lemma kbo_prec_kbo:
  "kbo_prec pr_strict pr_weak = kbo"
  (is "?l = ?r")
proof (intro ext)
  fix s t
  show "?l s t = ?r s t"
  proof (induct S\<equiv>pr_strict W\<equiv>pr_weak s t rule: kbo_prec.induct)
    case (1 s t)
    show ?case
    proof (cases "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> weight t \<le> weight s")
      case True
      show ?thesis
      proof (cases "weight t < weight s")
        note True' = True
        case True with True' show ?thesis by simp
      next
        case False
        show ?thesis
        proof (cases s)
          case (Var x) with True and False show ?thesis by auto
        next
          case (Fun f ss)
          show ?thesis
          proof (cases t)
            case (Var y) with True and False and Fun show ?thesis by auto
          next
            note Fun' = Fun
            case (Fun g ts)
            show ?thesis
            proof (cases "pr_strict (f, length ss) (g, length ts)")
              note True' = True
              case True with True' and False and Fun' and Fun show ?thesis by auto
            next
              note False' = False
              case False
              show ?thesis
              proof (cases "pr_weak (f, length ss) (g, length ts)")
                note False'' = False
                case False
                with True and False' and Fun' and Fun and False'' and False
                  show ?thesis by auto
              next
                note True' = True
                case True
                note IH = 1(1) [OF True' False' Fun' Fun False True _ _]
                have "lex_ext_unbounded ?l ss ts = lex_ext_unbounded ?r ss ts"
                  apply (rule lex_ext_unbounded_cong, simp, simp)
                  by (insert IH, simp)
                with True and True' and False' and Fun' and Fun and False
                  show ?thesis by simp
              qed
            qed
          qed
        qed
      qed
    next
      case False
      then show ?thesis by auto
    qed
  qed
qed

(* abbreviations for strict (S) and nonstrict (NS) KBO *)
abbreviation S where "S \<equiv> \<lambda> s t. fst (kbo s t)"
abbreviation NS where "NS \<equiv> \<lambda> s t. snd (kbo s t)"

end

locale admissible_weight_fun_prc = weight_fun_prc w w0 least scf pr_strict pr_weak
  for w w0 pr_strict pr_weak and least :: "'f \<Rightarrow> bool" and scf +
  assumes w0: "w (f, 0) \<ge> w0" "w0 > 0"
    and adm: "w (f, 1) = 0 \<Longrightarrow> pr_weak (f, 1) (g, n)"
    and least: "least f = (w (f, 0) = w0 \<and> (\<forall> g. w (g, 0) = w0 \<longrightarrow> pr_weak (g, 0) (f, 0)))"
    and scf: "i < n \<Longrightarrow> scf (f, n) i > 0"
    and pr_weak_refl [simp]: "pr_weak fn fn"
    and pr_weak_trans: "pr_weak fn gm \<Longrightarrow> pr_weak gm hk \<Longrightarrow> pr_weak fn hk"
    and pr_strict: "pr_strict fn gm = (pr_weak fn gm \<and> \<not> pr_weak gm fn)"
    and pr_SN: "SN {(fn, gm). pr_strict fn gm}"
begin


lemma weight_w0: "weight t \<ge> w0"
proof (induct t)
  case (Fun f ts)
  show ?case
  proof (cases ts)
    case Nil
    with w0(1) have "w0 \<le> w (f,length ts)" by auto
    thus ?thesis by auto
  next
    case (Cons s ss)
    then obtain i where i: "i < length ts" by auto
    from scf[OF this] have scf: "0 < scf (f,length ts) i" by auto
    then obtain n where scf: "scf (f,length ts) i = Suc n" by (auto elim: lessE)
    from id_take_nth_drop[OF i] i obtain bef aft where ts: "ts = bef @ ts ! i # aft" and ii: "length bef = i" by auto
    def tsi \<equiv> "ts ! i"
    note ts = ts[folded tsi_def]
    from i have tsi: "tsi \<in> set ts" unfolding tsi_def by auto
    from Fun[OF this] have w0: "w0 \<le> weight tsi" .
    show ?thesis using scf ii w0 unfolding ts
      by simp
  qed
qed simp

lemma weight_gt_0: "weight t > 0"
  using weight_w0 [of t] and w0 by arith

lemma weight_0 [iff]: "weight t = 0 \<longleftrightarrow> False"
  using weight_gt_0 [of t] by auto

lemma not_S_Var: "\<not> S (Var x) t"
  using weight_w0[of t] by (cases t, auto)

lemma S_imp_NS: "S s t \<Longrightarrow> NS s t"
proof (induct s t rule: kbo.induct)
  case (1 s t)
  from 1(2) have S: "S s t" .
  from S have w: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> weight t \<le> weight s" 
    by (auto split: if_splits)
  note S = S w
  note IH = 1(1)[OF w]
  show ?case
  proof (cases "weight t < weight s")
    case True
    with S show ?thesis by simp
  next
    case False
    note IH = IH[OF False]
    note S = S False
    from not_S_Var[of _ t] S
    obtain f ss where s: "s = Fun f ss" by (cases s, auto)
    note IH = IH[OF s]
    show ?thesis
    proof (cases t)
      case (Var x)
      from S show ?thesis by (auto, insert Var s, auto)
    next
      case (Fun g ts)
      note IH = IH[OF Fun]      
      let ?f = "(f,length ss)"
      let ?g = "(g,length ts)"
      let ?lex = "lex_ext_unbounded kbo ss ts"
      from S[simplified, unfolded s Fun] have disj: "pr_strict ?f ?g \<or> pr_weak ?f ?g \<and> fst ?lex" by (auto split: if_splits)
      show ?thesis
      proof (cases "pr_strict ?f ?g")
        case True
        thus ?thesis using S s Fun by auto
      next
        case False
        with disj have fg: "pr_weak ?f ?g" and lex: "fst ?lex" by auto
        note IH = IH[OF False fg]
        from lex have "fst (lex_ext kbo (length ss + length ts) ss ts)" 
          unfolding lex_ext_def Let_def by auto
        from lex_ext_stri_imp_nstri[OF this] have lex: "snd ?lex" 
          unfolding lex_ext_def Let_def by auto
        with False fg S s Fun show ?thesis by auto
      qed
    qed
  qed
qed

lemma NS_refl: "NS s s"
proof (induct s)
  case (Var x)
  show ?case by simp
next
  case (Fun f ss)
  have "snd (lex_ext kbo (length ss) ss ss)"
    by (rule all_nstri_imp_lex_nstri, insert Fun[unfolded set_conv_nth], auto)
  hence "snd (lex_ext_unbounded kbo ss ss)" unfolding lex_ext_def Let_def by simp
  thus ?case by auto 
qed

lemma pr_strict_irrefl: "\<not> pr_strict fn fn" 
  unfolding pr_strict by auto

lemma pr_strict_trans:
  "pr_strict fn gm \<Longrightarrow> pr_strict gm hk \<Longrightarrow> pr_strict fn hk"
  using pr_weak_trans
  unfolding pr_strict by blast

lemma S_irrefl: "\<not> S t t"
proof (induct t)
  case (Var x) then show ?case by (rule not_S_Var)
next
  case (Fun f ts)
  from pr_strict_irrefl have "\<not> pr_strict (f, length ts) (f, length ts)" .
  moreover
  { assume "fst (lex_ext_unbounded kbo ts ts)"
    then obtain i where "i < length ts" and "S (ts ! i) (ts ! i)"
      unfolding lex_ext_unbounded_iff by auto
    with Fun have False by auto }
  ultimately show ?case by auto
qed

lemma all_gt_0_listsum_map:
  assumes *: "\<And>x. f x > (0::nat)"
    and "x \<in> set xs" and "1 < length xs"
  shows "f x < (\<Sum>x\<leftarrow>xs. f x)"
  using assms
proof (induct xs)
  case (Cons x xs)
  then show ?case
    apply (auto)
    apply (cases xs, simp+)
    apply (cases xs, simp+)
    apply (case_tac list, simp+)
  done
qed simp

lemma weight_arg_strict:
  assumes "t \<in> set ts" and "1 < length ts" and "length ts \<le> n"
  shows "weight t < listsum (map weight (scf_list (scf (f, n)) ts))"
    (is "_ < ?sum")
proof -
  from all_gt_0_listsum_map [OF weight_gt_0 assms(1,2), of "\<lambda>x. x"]
    have "weight t < (\<Sum>t\<leftarrow>ts. weight t)" .
  also have "\<dots> \<le> ?sum"
    using listsum_scf_list [of ts "scf (f, n)"]
    using scf [of _ n f] and assms(3) by auto
  finally show ?thesis .
qed

lemma weight_arg_strict':
  assumes "t \<in> set ts" and "1 < length ts"
  shows "weight t < weight (Fun f ts)"
  using weight_arg_strict [OF assms order_refl, of f]
  unfolding weight_simp by arith

lemma weight_arg:
  assumes "t \<in> set ts"
  shows "weight t \<le> listsum (map weight (scf_list (scf (f, length ts)) ts))"
proof -
  from assms obtain ts1 ts2 where ts: "ts = ts1 @ t # ts2"
    unfolding in_set_conv_decomp by blast
  let ?ts = "ts1 @ t # ts2"
  have "length ts1 < length ?ts" by auto
  from scf [OF this, of f] obtain n
    where scf: "scf (f, length ?ts) (length ts1) = Suc n" by (auto elim: lessE)
  show ?thesis
    by (simp add: scf [simplified] ts)
qed

lemma weight_arg':
  assumes "t \<in> set ts"
  shows "weight t \<le> weight (Fun f ts)"
  using weight_arg [OF assms, of f] by auto

lemma vars_term_ms_SCF_arg:
  assumes "t \<in> set ts"
  shows "vars_term_ms (SCF t) \<le> vars_term_ms (SCF (Fun f ts))"
proof -
  from assms obtain ts1 ts2 where ts: "ts = ts1 @ t # ts2"
    unfolding in_set_conv_decomp by blast
  let ?ts = "ts1 @ t # ts2"
  have "length ts1 < length ?ts" by auto
  from scf [OF this] obtain n
    where scf: "scf (f, length ?ts) (length ts1) = Suc n" by (auto elim: lessE)
  obtain X where "vars_term_ms (SCF (Fun f ?ts)) = vars_term_ms (SCF t) + X"
    by (simp add: o_def scf [simplified], metis union_commute union_lcomm)
  then show ?thesis by (simp add: ts)
qed

lemma NS_Var_imp_eq_least: "NS (Var x) t \<Longrightarrow> t = Var x \<or> (\<exists> f. t = Fun f [] \<and> least f)"
  by (cases t, insert weight_w0[of t], auto split: if_splits)

lemma S_mono_one:
  assumes S: "S s t"
  shows "S (Fun f (ss1 @ s # ss2)) (Fun f (ss1 @ t # ss2))"
proof -
  let ?ss = "ss1 @ s # ss2"
  let ?ts = "ss1 @ t # ss2"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  from S have w: "weight t \<le> weight s" and v: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s)" 
    by (auto split: if_splits)
  have v': "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF ?s)" using multiset_of_replicate_mono[OF v] by simp
  have w': "weight ?t \<le> weight ?s" using listsum_replicate_mono[OF w] by simp
  have lex: "fst (lex_ext_unbounded kbo ?ss ?ts)"
    unfolding lex_ext_unbounded_iff fst_conv
    by (rule disjI1, rule exI[of _ "length ss1"], insert S NS_refl, auto simp del: kbo.simps simp: nth_append)    
  show ?thesis using v' w' lex by simp
qed

lemma S_ctxt: "S s t \<Longrightarrow> S (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using one_imp_ctxt_closed [of "{(x, y). S x y}", to_pred, OF S_mono_one, THEN ctxt_closedE, to_pred]
  by blast

lemma NS_mono_one:
  assumes NS: "NS s t" shows "NS (Fun f (ss1 @ s # ss2)) (Fun f (ss1 @ t # ss2))"
proof -
  let ?ss = "ss1 @ s # ss2"
  let ?ts = "ss1 @ t # ss2"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  from NS have w: "weight t \<le> weight s" and v: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s)" 
    by (auto split: if_splits)
  have v': "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF ?s)" using multiset_of_replicate_mono[OF v] by simp
  have w': "weight ?t \<le> weight ?s" using listsum_replicate_mono[OF w] by simp
  have lex: "snd (lex_ext_unbounded kbo ?ss ?ts)"
    unfolding lex_ext_unbounded_iff snd_conv
  proof (intro disjI2 conjI allI impI)
    fix i
    assume "i < length (ss1 @ t # ss2)"
    thus "NS (?ss ! i) (?ts ! i)" using NS NS_refl
      by (cases "i = length ss1", auto simp del: kbo.simps simp: nth_append)
  qed simp
  show ?thesis using v' w' lex by simp
qed

lemma NS_ctxt: "NS s t \<Longrightarrow> NS (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using one_imp_ctxt_closed [of "{(x, y). NS x y}", to_pred, OF NS_mono_one, THEN ctxt_closedE, to_pred]
  by blast

(* main auxiliary lemma for proving subterm property *)
lemma kbo_supt_one: "NS s t \<Longrightarrow> S (Fun f (bef @ s # aft)) t"
proof (induct t arbitrary: f s bef aft)
  case (Var x)
  note NS = this
  let ?ss = "bef @ s # aft"
  let ?t = "Var x"
  have "length bef < length ?ss" by auto
  from scf[OF this, of f] obtain n where scf:"scf (f,length ?ss) (length bef) = Suc n" by (auto elim: lessE)
  obtain X where "vars_term_ms (SCF (Fun f ?ss)) = vars_term_ms (SCF s) + X"
    by (simp add: o_def scf[simplified], metis mset_le_add_left union_commute union_lcomm)
  hence vs: "vars_term_ms (SCF s) \<le> vars_term_ms (SCF (Fun f ?ss))" by simp
  from NS have vt: "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF s)" by (auto split: if_splits)
  from vt vs have v: "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF (Fun f ?ss))" by (rule mset_le_trans)
  from weight_w0[of "Fun f ?ss"] v show ?case by simp
next
  case (Fun g ts f s bef aft)
  let ?t = "Fun g ts"
  let ?ss = "bef @ s # aft"
  note NS = Fun(2)
  note IH = Fun(1)
  have "length bef < length ?ss" by auto
  from scf[OF this, of f] obtain n where scff:"scf (f,length ?ss) (length bef) = Suc n" by (auto elim: lessE)
  note scff = scff[simplified]
  obtain X where "vars_term_ms (SCF (Fun f ?ss)) = vars_term_ms (SCF s) + X"
    by (simp add: o_def scff, metis mset_le_add_left union_commute union_lcomm)
  hence vs: "vars_term_ms (SCF s) \<le> vars_term_ms (SCF (Fun f ?ss))" by simp
  have ws: "weight s \<le> listsum (map weight (scf_list (scf (f,length ?ss)) ?ss))" 
    by (simp add: scff)
  from NS have wt: "weight ?t \<le> weight s" and 
    vt: "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF s)" by (auto split: if_splits)
  from ws wt have w: "weight ?t \<le> listsum (map weight (scf_list (scf (f,length ?ss)) ?ss))" by simp
  from vt vs have v: "vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF (Fun f ?ss))" by auto
  hence v': "(vars_term_ms (SCF ?t) \<le> vars_term_ms (SCF (Fun f ?ss))) = True" by simp
  show ?case
  proof (cases "weight ?t = weight (Fun f ?ss)")
    case False
    with w v show ?thesis by auto
  next
    case True
    from wt[unfolded True] weight_gt_0[of s]
    have wf: "w (f,length ?ss) = 0"
      and lsum: "listsum (map weight (scf_list (scf (f,length ?ss)) bef)) = 0" 
         "listsum (map weight (scf_list (\<lambda> i. (scf (f,length ?ss) (Suc (length bef) + i))) aft)) = 0"
         and n: "n = 0"
         by (auto simp: scff)
    have "listsum (map weight bef) \<le> listsum (map weight (scf_list (scf (f,length ?ss)) bef))"
      by (rule listsum_scf_list, rule scf, auto)
    with lsum(1) have "listsum (map weight bef) = 0" by arith
    hence bef: "bef = []" using weight_gt_0[of "hd bef"] by (cases bef, auto)
    have "listsum (map weight aft) \<le> listsum (map weight (scf_list (\<lambda> i. (scf (f,length ?ss) (Suc (length bef) + i))) aft))"
      by (rule listsum_scf_list, rule scf, auto)
    with lsum(2) have "listsum (map weight aft) = 0" by arith
    hence aft: "aft = []" using weight_gt_0[of "hd aft"] by (cases aft, auto)
    note scff = scff[unfolded bef aft n, simplified]
    from bef aft
    have ba: "bef @ s # aft = [s]" by simp
    with wf have wf: "w (f,1) = 0" by auto
    from wf have wst: "weight s = weight ?t" using scff unfolding True[unfolded ba] 
      by (simp add: scf_list_def)
    let ?g = "(g,length ts)"
    let ?f = "(f,1)"
    show ?thesis
    proof (cases "pr_strict ?f ?g")
      case True
      with w v show ?thesis unfolding ba by simp
    next
      case False
      note admf = adm[OF wf]
      from admf have pg: "pr_weak ?f ?g" .
      from pg False[unfolded pr_strict] have "pr_weak ?g ?f" by auto
      from pr_weak_trans[OF this admf] have g: "\<And> h k. pr_weak ?g (h,k)" .
      show ?thesis
      proof (cases ts)
        case Nil
        have "fst (lex_ext_unbounded kbo [s] ts)" 
          unfolding Nil lex_ext_unbounded_iff by auto
        with pg w v show ?thesis unfolding ba by simp
      next
        case (Cons t tts)
        {
          fix x
          assume s: "s = Var x"          
          from NS_Var_imp_eq_least[OF NS[unfolded s Cons]] have False by auto
        }
        then obtain h ss where s: "s = Fun h ss" by (cases s, auto)
        from NS wst g[of h "length ss"] pr_strict[of "(h,length ss)" "(g,length ts)"] have lex: "snd (lex_ext_unbounded kbo ss ts)"
          unfolding s by (auto split: if_splits)
        from lex obtain s0 sss where ss: "ss = s0 # sss" unfolding Cons lex_ext_unbounded_iff snd_conv by (cases ss, auto)
        from lex[unfolded ss Cons] have "S s0 t \<or> NS s0 t"
          by (cases "kbo s0 t", simp add: lex_ext_unbounded.simps del: kbo.simps split: if_splits)
        with S_imp_NS[of s0 t] have "NS s0 t" by blast
        from IH[OF _ this, of h Nil sss] have S: "S s t" unfolding Cons s ss by simp
        have "fst (lex_ext_unbounded kbo [s] ts)" unfolding Cons
          unfolding lex_ext_unbounded_iff fst_conv
          by (rule disjI1[OF exI[of _ 0]], insert S, auto simp del: kbo.simps)
        hence lex: "fst (lex_ext_unbounded kbo [s] ts) = True" by simp
        note all = lex wst[symmetric] S pg scff v'
        note all = all[unfolded ba, unfolded s ss Cons]
        {
          fix t
          have "weight (Fun f [t]) = weight t"
            using wf scff by (simp add: scf_list_def)
        } note w = this
        show ?thesis unfolding ba unfolding s ss Cons
          unfolding kbo.simps[of "Fun f [Fun h (s0 # sss)]"]
          unfolding all w using all by simp
      qed
    qed
  qed
qed

lemma kbo_supt:
  assumes supt: "s \<rhd> t"
  shows "S s t"
proof -
  from supt obtain C where s: "s = C\<langle>t\<rangle>" and C: "C \<noteq> \<box>" by auto
  show ?thesis unfolding s using C
  proof (induct C arbitrary: t)
    case Hole thus ?case by simp
  next
    case (More f bef C aft t)
    show ?case
    proof (cases "C = \<box>")
      case True
      from kbo_supt_one[OF NS_refl, of f bef t aft] show ?thesis unfolding True by simp
    next
      case False
      from kbo_supt_one[OF S_imp_NS[OF More(1)[OF False]], of f bef t aft]
      show ?thesis by simp
    qed
  qed
qed

lemma kbo_supteq: assumes supt: "s \<unrhd> t" shows "NS s t"
  using S_imp_NS[OF kbo_supt[of s t]] NS_refl[of s] using supt[unfolded subterm.le_less]
  by blast

lemma S_subterm_one:
  assumes "t \<in> set ts"
  shows "S (Fun f ts) t"
proof -
  from `t \<in> set ts` have "Fun f ts \<rhd> t" by auto
  with kbo_supt show ?thesis by blast
qed

lemma NS_all_least: assumes l: "least f"
  shows "NS t (Fun f [])"
proof (induct t)
  case (Var x)
  show ?case using l[unfolded least] l 
    by auto
next
  case (Fun g ts)
  show ?case
  proof (cases ts)
    case (Cons s ss)
    with Fun[of s] have "NS s (Fun f [])" by auto
    from S_imp_NS[OF kbo_supt_one[OF this, of g Nil ss]] show ?thesis unfolding Cons by simp
  next
    case Nil
    from weight_w0[of "Fun g []"] have w: "weight (Fun g []) \<ge> weight (Fun f [])"
      using l[unfolded least] by auto
    from lex_ext_least_1
    have "snd (lex_ext kbo 0 [] [])" .
    hence lex: "snd (lex_ext_unbounded kbo [] [])" unfolding lex_ext_def Let_def by simp
    thus ?thesis using w l[unfolded least] unfolding Fun Nil by (auto simp: empty_le)
  qed
qed


lemma weight_subst: "weight (t \<cdot> \<sigma>) = weight t + mset_sum (Multiset2.map (\<lambda> x. weight (\<sigma> x) - w0) (vars_term_ms (SCF t)))"
proof (induct t)
  case (Var x)
  show ?case using weight_w0[of "\<sigma> x"] by auto
next
  case (Fun f ts)
  let ?ts = "scf_list (scf (f, length ts)) ts"
  def sts \<equiv> ?ts
  have id: "map (\<lambda> t. weight (t \<cdot> \<sigma>)) ?ts = map (\<lambda> t. weight t + mset_sum (Multiset2.map (\<lambda> x. weight (\<sigma> x) - w0) (vars_term_ms (scf_term scf t)))) ?ts"
    by (rule map_cong[OF refl Fun], insert scf_list_subset[of _ ts], auto)
  show ?case
    by (simp add: o_def id, unfold sts_def[symmetric], induct sts, auto)
qed

lemma weight_stable_le: assumes ws: "weight s \<le> weight t" and 
  vs: "vars_term_ms (SCF s) \<le> vars_term_ms (SCF t)"
  shows "weight (s \<cdot> \<sigma>) \<le> weight (t \<cdot> \<sigma>)"
proof -
  from vs[unfolded mset_le_exists_conv] obtain u where vt: "vars_term_ms (SCF t) = vars_term_ms (SCF s) + u" ..
  show ?thesis unfolding weight_subst vt using ws by auto
qed

lemma weight_stable_lt: assumes ws: "weight s < weight t" and 
  vs: "vars_term_ms (SCF s) \<le> vars_term_ms (SCF t)"
  shows "weight (s \<cdot> \<sigma>) < weight (t \<cdot> \<sigma>)"
proof -
  from vs[unfolded mset_le_exists_conv] obtain u where vt: "vars_term_ms (SCF t) = vars_term_ms (SCF s) + u" ..
  show ?thesis unfolding weight_subst vt using ws by auto
qed

lemma kbo_stable:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "NS s t"
  shows "(S s t \<longrightarrow> S (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)) \<and> NS (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" (is "?P s t")
  using assms
proof (induct s arbitrary: t)
  case (Var y t)
    hence not: "\<not> S (Var y) t" using not_S_Var[of y t] by auto
    from NS_Var_imp_eq_least[OF Var]
    have "t = Var y \<or> (\<exists> f. t = Fun f [] \<and> least f)" by simp
    then obtain f where "t = Var y \<or> t = Fun f [] \<and> least f" by auto
    hence "NS (Var y \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    proof
      assume "t = Var y"
      thus ?thesis using NS_refl[of "t \<cdot> \<sigma>"] by auto
    next
      assume "t = Fun f [] \<and> least f"
      with NS_all_least[of f "Var y \<cdot> \<sigma>"] show ?thesis by auto
    qed    
    with not show ?case by blast
  next
    case (Fun f ss t) 
    note NS = Fun(2)
    note IH = Fun(1)
    let ?s = "Fun f ss"
    def s \<equiv> ?s
    let ?ss = "map (\<lambda> s. s \<cdot> \<sigma>) ss"
    from NS have v: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF ?s)" and w: "weight t \<le> weight ?s"
      by (auto split: if_splits)
    from weight_stable_le[OF w v] have w\<sigma>: "weight (t \<cdot> \<sigma>) \<le> weight (?s \<cdot> \<sigma>)" by auto
    from vars_term_ms_subst_mono[OF v, of "\<lambda> x. SCF (\<sigma> x)"] have v\<sigma>: "vars_term_ms (SCF (t \<cdot> \<sigma>)) \<le> vars_term_ms (SCF (?s \<cdot> \<sigma>))" 
      unfolding scf_term_subst .
    show ?case
    proof (cases "weight (t \<cdot> \<sigma>) < weight (?s \<cdot> \<sigma>)")
      case True
      with v\<sigma> show ?thesis by auto
    next
      case False
      with weight_stable_lt[OF _ v, of \<sigma>] w have w: "weight t = weight ?s" by arith
      show ?thesis
      proof (cases t)
        case (Var y)
        from set_of_mono[OF v, folded s_def]
        have "y \<in> vars_term (SCF s)" unfolding Var by (auto simp: o_def)
        also have "\<dots> \<subseteq> vars_term s" by (rule vars_term_scf_subset)
        finally have "y \<in> vars_term s" by auto
        from supteq_Var[OF this] have "?s \<rhd> Var y" unfolding s_def Fun by auto 
        from kbo_supt[OF supt_stable[OF this]] have S: "S (?s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" unfolding Var .
        from S_imp_NS[OF S] S show ?thesis by auto
      next
        case (Fun g ts) note t = this
        let ?f = "(f,length ss)"
        let ?g = "(g,length ts)"
        let ?ts = "map (\<lambda> s. s \<cdot> \<sigma>) ts"
        show ?thesis
        proof (cases "pr_strict ?f ?g")
          case True 
          hence S: "S (?s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" using w\<sigma> v\<sigma> unfolding t by simp
          from S S_imp_NS[OF S] show ?thesis by simp
        next
          case False note prec = this
          show ?thesis
          proof (cases "pr_weak ?f ?g")
            case False
            with v w prec have "\<not> NS ?s t" unfolding t by (auto simp del: vars_term_ms.simps)
            with NS show ?thesis by blast
          next
            case True
            from v w have "vars_term_ms (SCF t) \<le> vars_term_ms (SCF ?s) \<and> weight t \<le> weight ?s" "\<not> weight t < weight ?s" by auto
            {
              fix i
              assume i: "i < length ss" "i < length ts"
              and S: "S (ss ! i) (ts ! i)"
              have "S (map (\<lambda>s. s \<cdot> \<sigma>) ss ! i) (map (\<lambda>s. s \<cdot> \<sigma>) ts ! i)"
                using IH[OF _ S_imp_NS[OF S]] S i unfolding set_conv_nth by (force simp del: kbo.simps)
            } note IH_S = this
            {
              fix i
              assume i: "i < length ss" "i < length ts"
              and NS: "NS (ss ! i) (ts ! i)"
              have "NS (map (\<lambda>s. s \<cdot> \<sigma>) ss ! i) (map (\<lambda>s. s \<cdot> \<sigma>) ts ! i)"
                using IH[OF _ NS] i unfolding set_conv_nth by (force simp del: kbo.simps)
            } note IH_NS = this
            {
              assume "S ?s t"
              with prec v w True have lex: "fst (lex_ext_unbounded kbo ss ts)"
                unfolding s_def t by simp
              have "fst (lex_ext_unbounded kbo ?ss ?ts)"
                by (rule lex_ext_unbounded_map_S[OF _ _ lex], insert IH_NS IH_S, blast+)
                with v\<sigma> w\<sigma> prec True have "S (?s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
                unfolding t by auto
            }
            moreover 
            {
              from NS prec v w True have lex: "snd (lex_ext_unbounded kbo ss ts)"
                unfolding t by simp
              have "snd (lex_ext_unbounded kbo ?ss ?ts)"
                by (rule lex_ext_unbounded_map_NS[OF _ _ lex], insert IH_S IH_NS, blast)
              with v\<sigma> w\<sigma> prec True have "NS (?s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
                unfolding t by auto
            }
            ultimately show ?thesis by auto
        qed
      qed
    qed
  qed
qed

lemma S_subst:
  "S s t \<Longrightarrow> S (s \<cdot> (\<sigma> :: ('f, 'v) subst)) (t \<cdot> \<sigma>)"
  using kbo_stable[OF S_imp_NS, of s t \<sigma>] by auto

lemma NS_subst: "NS s t \<Longrightarrow> NS (s \<cdot> (\<sigma> :: ('f,'v)subst)) (t \<cdot> \<sigma>)" using kbo_stable[of s t \<sigma>] by auto

(* the following two properties on least are used to prove transitivity *)
lemma not_S_least: assumes l: "least f"
  shows "\<not> S (Fun f []) t"
proof (cases t)
  case (Var x)
  show ?thesis unfolding Var by simp
next
  case (Fun g ts)
  show ?thesis unfolding Fun
  proof 
    assume S: "S (Fun f []) (Fun g ts)"
    from S[unfolded Fun, simplified]
    have w: "w (g,length ts) + listsum (map weight (scf_list (scf (g,length ts)) ts)) \<le> weight (Fun f [])"
      by (auto split: if_splits)
    show False
    proof (cases ts)
      case Nil
      with w have "w (g,0) \<le> weight (Fun f [])" by simp
      also have "weight (Fun f []) \<le> w0" using l[unfolded least] by simp
      finally have g: "w (g,0) = w0" using  w0(1)[of g] by auto
      with w Nil l[unfolded least] have gf: "w (g,0) = w (f,0)" by simp
      with S have p: "pr_weak (f,0) (g,0)" unfolding Nil 
        by (simp split: if_splits add: pr_strict)
      with l[unfolded least, THEN conjunct2, rule_format, OF g] have p2: "pr_weak (g,0) (f,0)" by auto
      from p p2 gf S have "fst (lex_ext_unbounded kbo [] ts)" unfolding Nil
        by (auto simp: pr_strict)
      thus False unfolding lex_ext_unbounded_iff by auto
    next
      case (Cons s ss)
      hence ts: "ts = [] @ s # ss" by auto
      from scf[of 0 "length ts" g] obtain n where scff: "scf (g,length ts) 0 = Suc n" unfolding Cons by (auto elim: lessE)
      let ?e = "listsum (map weight (
          scf_list (\<lambda>i. scf (g, Suc (length ss)) (Suc i)) ss
          ))"
      have "w0 + listsum (map weight (replicate n s)) \<le> weight s + listsum (map weight (replicate n s))"
        using weight_w0[of s] by auto 
      also have "\<dots> = listsum (map weight (replicate (scf (g,length ts) 0) s))" unfolding scff by simp
      also have "w (g,length ts) + \<dots> + ?e \<le> w0" using w l[unfolded least] unfolding ts scf_list_bef_i_aft by auto
      finally have "w0 + listsum (map weight (replicate n s)) + w (g,length ts) + ?e \<le> w0" by arith
      hence wg: "w (g,length ts) = 0" and null: "?e = 0" "listsum (map weight (replicate n s)) = 0" by auto
      from null(2) weight_gt_0[of s] have n: "n = 0" by (cases n, auto)
      have "listsum (map weight ss) \<le> ?e"
        by (rule listsum_scf_list, rule scf, auto)
      from this[unfolded null] weight_gt_0[of "hd ss"] have ss: "ss = []" by (cases ss, auto)
      with Cons have ts: "ts = [s]" by simp
      note scff = scff[unfolded ts n, simplified]
      from wg ts have wg: "w (g,1) = 0" by auto
      from adm[OF wg, rule_format, of f] have "pr_weak (g,1) (f,0)" by auto
      with S[unfolded Fun ts] l[unfolded least] weight_w0[of s] scff
      have "fst (lex_ext_unbounded kbo [] [s])" 
        by (auto split: if_splits simp: scf_list_def pr_strict)
      thus ?thesis unfolding lex_ext_unbounded_iff by auto
    qed
  qed
qed

lemma NS_least_least: assumes l: "least f"
  and NS: "NS (Fun f []) t"
  shows "\<exists> g. t = Fun g [] \<and> least g"
proof (cases t)
  case (Var x)
  show ?thesis using NS unfolding Var by simp
next
  case (Fun g ts)
  from NS[unfolded Fun, simplified]
  have w: "w (g,length ts) + listsum (map weight (scf_list (scf (g,length ts)) ts)) \<le> weight (Fun f [])"
    by (auto split: if_splits)
  show ?thesis
  proof (cases ts)
    case Nil
    with w have "w (g,0) \<le> weight (Fun f [])" by simp
    also have "weight (Fun f []) \<le> w0" using l[unfolded least] by simp
    finally have g: "w (g,0) = w0" using  w0(1)[of g] by auto
    with w Nil l[unfolded least] have gf: "w (g,0) = w (f,0)" by simp
    with NS[unfolded Fun] have p: "pr_weak (f,0) (g,0)" unfolding Nil 
      by (simp split: if_splits add: pr_strict)
    have least: "least g" unfolding least
    proof (rule conjI[OF g], intro allI)
      fix h
      from l[unfolded least] have "w (h,0) = w0 \<longrightarrow> pr_weak (h,0) (f,0)" by blast
      with pr_weak_trans p show "w (h,0) = w0 \<longrightarrow> pr_weak (h,0) (g,0)" by blast
    qed
    show ?thesis 
      by (rule exI[of _ g], unfold Fun Nil, insert least, auto)
  next
    case (Cons s ss)
    hence ts: "ts = [] @ s # ss" by auto
    from scf[of 0 "length ts" g] obtain n where scff: "scf (g,length ts) 0 = Suc n" unfolding Cons by (auto elim: lessE)
    let ?e = "listsum (map weight (
        scf_list (\<lambda>i. scf (g, Suc (length ss)) (Suc i)) ss
      ))"
    have "w0 + listsum (map weight (replicate n s)) \<le> weight s + listsum (map weight (replicate n s))"
      using weight_w0[of s] by auto 
    also have "\<dots> = listsum (map weight (replicate (scf (g,length ts) 0) s))" unfolding scff by simp
    also have "w (g,length ts) + \<dots> + ?e \<le> w0" using w l[unfolded least] unfolding ts scf_list_bef_i_aft by auto
    finally have "w0 + listsum (map weight (replicate n s)) + w (g,length ts) + ?e \<le> w0" by arith
    hence wg: "w (g,length ts) = 0" and null: "?e = 0" "listsum (map weight (replicate n s)) = 0" by auto
    from null(2) weight_gt_0[of s] have n: "n = 0" by (cases n, auto)
    have "listsum (map weight ss) \<le> ?e"
      by (rule listsum_scf_list, rule scf, auto)
    from this[unfolded null] weight_gt_0[of "hd ss"] have ss: "ss = []" by (cases ss, auto)
    with Cons have ts: "ts = [s]" by simp
    note scff = scff[unfolded ts n, simplified]
    from wg ts have wg: "w (g,1) = 0" by auto
    from adm[OF wg, rule_format, of f] have "pr_weak (g,1) (f,0)" by auto
    with NS[unfolded Fun ts] l[unfolded least] weight_w0[of s] scff
    have "snd (lex_ext_unbounded kbo [] [s])" 
      by (auto split: if_splits simp: scf_list_def pr_strict)
    thus ?thesis unfolding lex_ext_unbounded_iff snd_conv by auto
  qed
qed

lemma kbo_trans: "(S s t \<longrightarrow> NS t u \<longrightarrow> S s u) \<and> (NS s t \<longrightarrow> S t u \<longrightarrow> S s u) \<and> (NS s t \<longrightarrow> NS t u \<longrightarrow> NS s u)" (is "?P s t u")
proof (induct s arbitrary: t u)
  case (Var x t u)
  from not_S_Var[of x t] have nS: "\<not> S (Var x) t" .
  show ?case
  proof (cases "NS (Var x) t")
    case False
    with nS show ?thesis by blast
  next
    case True
    from NS_Var_imp_eq_least[OF this] obtain f where 
      "t = Var x \<or> t = Fun f [] \<and> least f" by blast
    thus ?thesis
    proof 
      assume "t = Var x"
      thus ?thesis using nS by blast
    next
      assume "t = Fun f [] \<and> least f"
      hence t: "t = Fun f []" and least: "least f" by auto
      from not_S_least[OF least] have nS': "\<not> S t u" unfolding t .
      show ?thesis
      proof (cases "NS t u")
        case True
        with NS_least_least[OF least, of u] t obtain h where 
          u: "u = Fun h []" and least: "least h" by auto
        from NS_all_least[OF least] have NS: "NS (Var x) u" unfolding u .
        with nS nS' show ?thesis by blast
      next
        case False
        with S_imp_NS[of t u] show ?thesis by blast
      qed
    qed
  qed   
next
  case (Fun f ss t u) note IH = this
  let ?s = "Fun f ss"
  show ?case
  proof (cases "NS ?s t")
    case False
    with S_imp_NS[of ?s t] show ?thesis by blast
  next
    case True note st = this
    hence vst: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF ?s)" and wst: "weight t \<le> weight ?s"
      by (auto split: if_splits)
    show ?thesis
    proof (cases "NS t u")
      case False
      with S_imp_NS[of t u] show ?thesis by blast
    next
      case True note tu = this
      hence vtu: "vars_term_ms (SCF u) \<le> vars_term_ms (SCF t)" and wtu: "weight u \<le> weight t"
        by (auto split: if_splits)
      from vst vtu have v: "vars_term_ms (SCF u) \<le> vars_term_ms (SCF ?s)" by simp
      from wst wtu have w: "weight u \<le> weight ?s" by simp
      show ?thesis
      proof (cases "weight u < weight ?s")
        case True
        with v show ?thesis by auto
      next
        case False
        with wst wtu have wst: "weight t = weight ?s" and wtu: "weight u = weight t" and w: "weight u = weight ?s" by arith+                
        show ?thesis
        proof (cases u)
          case (Var z)
          with v w show ?thesis by auto
        next
          case (Fun h us) note u = this
          show ?thesis
          proof (cases t)
            case (Fun g ts) note t = this
            let ?f = "(f,length ss)"
            let ?g = "(g,length ts)"
            let ?h = "(h,length us)"
            from st t wst have fg: "pr_weak ?f ?g" by (simp split: if_splits add: pr_strict)
            from tu t u wtu have gh: "pr_weak ?g ?h" by (simp split: if_splits add: pr_strict)
            from pr_weak_trans[OF fg gh] have fh: "pr_weak ?f ?h" .          
            show ?thesis
            proof (cases "pr_strict ?f ?h")
              case True
              with w v u show ?thesis by auto
            next
              case False
              let ?lex = "lex_ext_unbounded kbo"
              from False fh have hf: "pr_weak ?h ?f" unfolding pr_strict by auto
              from pr_weak_trans[OF hf fg] have hg: "pr_weak ?h ?g" .
              from hg have gh2: "\<not> pr_strict ?g ?h" unfolding pr_strict by auto
              from pr_weak_trans[OF gh hf] have gf: "pr_weak ?g ?f" .
              from gf have fg2: "\<not> pr_strict ?f ?g" unfolding pr_strict by auto
              from st t wst fg2 have st: "snd (?lex ss ts)"
                by (auto split: if_splits)
              from tu t u wtu gh2 have tu: "snd (?lex ts us)" 
                by (auto split: if_splits)
              {
                fix s t u
                assume "s \<in> set ss"
                from IH[OF this, of t u]
                have "(NS s t \<and> S t u \<longrightarrow> S s u) \<and> 
                  (S s t \<and> NS t u \<longrightarrow> S s u) \<and>
                  (NS s t \<and> NS t u \<longrightarrow> NS s u) \<and> 
                  (S s t \<and> S t u \<longrightarrow> S s u)"
                  using S_imp_NS[of s t] by blast
              } note IH = this
              let ?b = "length ss + length ts + length us"
              note lex = lex_ext_compat[of ss ts us kbo ?b, OF IH]
              let ?lexb = "lex_ext kbo ?b"
              note conv = lex_ext_def Let_def
              from st have st: "snd (?lexb ss ts)" unfolding conv by simp
              from tu have tu: "snd (?lexb ts us)" unfolding conv by simp
              from lex st tu have su: "snd (?lexb ss us)" by blast
              hence su: "snd (?lex ss us)" unfolding conv by simp
              from w v u su fh have NS: "NS ?s u" by simp
              {
                assume st: "S ?s t"
                with t wst fg fg2 have st: "fst (?lex ss ts)"
                  by (auto split: if_splits)
                hence st: "fst (?lexb ss ts)" unfolding conv by simp
                from lex st tu have su: "fst (?lexb ss us)" by blast
                hence su: "fst (?lex ss us)" unfolding conv by simp
                from w v u su fh have S: "S ?s u" by simp
              } note S_left = this
              {
                assume tu: "S t u"
                with t u wtu gh2 have tu: "fst (?lex ts us)"
                  by (auto split: if_splits)
                hence tu: "fst (?lexb ts us)" unfolding conv by simp
                from lex st tu have su: "fst (?lexb ss us)" by blast
                hence su: "fst (?lex ss us)" unfolding conv by simp
                from w v u su fh have S: "S ?s u" by simp
              } note S_right = this
              from NS S_left S_right show ?thesis by blast
            qed
          next
            case (Var x) note t = this
            from tu weight_w0[of u] have least: "least h" and u: "u = Fun h []" unfolding t u
              by (auto split: if_splits)
            from NS_all_least[OF least] have NS: "NS ?s u" unfolding u .
            from not_S_Var have nS': "\<not> S t u" unfolding t .
            show ?thesis
            proof (cases "S ?s t")
              case False
              with nS' NS show ?thesis by blast
            next
              case True
              hence "vars_term_ms (SCF t) \<le> vars_term_ms (SCF ?s)" 
                by (auto split: if_splits)
              from set_of_mono[OF this, unfolded set_of_vars_term_ms t]
              have "x \<in> vars_term (SCF ?s)" by simp
              also have "\<dots> \<subseteq> vars_term ?s" by (rule vars_term_scf_subset)
              finally obtain s sss where ss: "ss = s # sss" by (cases ss, auto)
              from kbo_supt_one[OF NS_all_least[OF least, of s], of f Nil sss]
              have "S ?s u" unfolding ss u by simp
              with NS show ?thesis by blast
            qed
          qed
        qed 
      qed
    qed
  qed
qed

lemma S_trans: "S s t \<Longrightarrow> S t u \<Longrightarrow> S s u" using S_imp_NS[of s t] kbo_trans[of s t u] by blast
lemma NS_trans: "NS s t \<Longrightarrow> NS t u \<Longrightarrow> NS s u" using kbo_trans[of s t u] by blast
lemma NS_S_compat: "NS s t \<Longrightarrow> S t u \<Longrightarrow> S s u" using kbo_trans[of s t u] by blast
lemma S_NS_compat: "S s t \<Longrightarrow> NS t u \<Longrightarrow> S s u" using kbo_trans[of s t u] by blast


section {* strong normalization via Kruskal *}

lemma pr_strict_reflclp_imp_pr_weak:
  assumes "pr_strict\<^sup>=\<^sup>= x y" shows "pr_weak x y"
  using assms by (auto simp: pr_strict)

lemma transp_on_pr_strict: "transp_on pr_strict UNIV"
  using pr_strict_trans unfolding transp_on_def by blast

lemma irreflp_on_pr_strict: "irreflp_on pr_strict UNIV"
  using pr_strict_irrefl unfolding irreflp_on_def by blast

lemma po_on_pr_strict: "po_on pr_strict\<inverse>\<inverse> UNIV"
proof
  show "irreflp_on pr_strict\<inverse>\<inverse> UNIV"
    by (simp) (rule irreflp_on_pr_strict)
next
  show "transp_on pr_strict\<inverse>\<inverse> UNIV"
    by (simp) (rule transp_on_pr_strict)
qed

lemma wfp_on_pr_strict: "wfp_on pr_strict\<inverse>\<inverse> (UNIV::('f \<times> nat) set)"
  using pr_SN unfolding wfp_on_SN_conv by auto

lemma ground_SCF:
  assumes "ground t"
  shows "ground (SCF t)"
  using assms
proof (induct t)
  case (Fun f ts)
  thus ?case using scf_list_subset[of _ ts] by auto
qed simp

lemma ground_imp_vars_term_ms_SCF_empty:
  assumes "ground t"
  shows "vars_term_ms (SCF t) = {#}"
  unfolding ground_vars_term_ms_empty[symmetric]
  by (rule ground_SCF [OF assms])

text {*Lemmas for \emph{simple weights} (i.e., without subterm coefficient functions).*}
context
  assumes scf': "i < n \<Longrightarrow> scf (f, n) i \<le> 1"
begin

lemma scf_1:
  assumes "i < n"
  shows "scf (f, n) i = 1"
  using scf [OF assms] and scf' [OF assms]
  by (metis (full_types) One_nat_def Suc_lessI not_less)

declare weight_simp [simp del]

lemma weight_simp' [simp]:
  "weight (Fun f ts) = w (f, length ts) + (\<Sum>t\<leftarrow>ts. weight t)"
proof -
  def n \<equiv> "length ts"
  have *: "zip (map weight ts) (map (scf (f, n)) [0..<n]) = map (\<lambda>t. (weight t, 1)) ts"
    (is "?xs = ?ys")
  proof (rule nth_equalityI)
    show length_eq: "length ?xs = length ?ys" by (simp add: n_def)
    show "\<forall>i<length ?xs. ?xs ! i = ?ys ! i"
    proof (intro allI impI)
      fix i
      assume "i < length ?xs"
      then have *: "i < length (map weight ts)" "i < length (map (scf (f, n)) [0..<n])"
        and "i < n"
        by auto
      then have len': "i < length [0..<n]" by (simp add: nth_def)
      from `i < n` have ith: "[0..<n] ! i = i" by simp
      have map_ith: "map (scf (f, n)) [0..<n] ! i = 1"
        using scf_1 [OF `i < n`] by (simp add: ith nth_map [OF len'])
      show "?xs ! i = ?ys ! i"
        unfolding nth_zip [OF *] map_ith
        using `i < n` unfolding n_def by simp
    qed
  qed
  have "(\<Sum>(ti, i)\<leftarrow>zip ts [0..<n]. weight ti * scf (f, n) i) =
    (\<Sum>tii\<leftarrow>zip ts [0..<n]. weight (fst tii) * scf (f, n) (snd tii))"
    unfolding split_def ..
  also have "\<dots> = (\<Sum>wic\<leftarrow>zip (map weight ts) (map (scf (f, n)) [0..<n]). fst wic * snd wic)"
    by (simp add: zip_map_map split_def o_def)
  also have "\<dots> = (\<Sum>wic\<leftarrow>map (\<lambda>t. (weight t, 1)) ts. fst wic * snd wic)" unfolding * ..
  also have "\<dots> = (\<Sum>wic\<leftarrow>map (\<lambda>t. (weight t, 1::nat)) ts. fst wic * 1)" by (induct ts) auto
  also have "\<dots> = (\<Sum>t\<leftarrow>ts. weight t)" by (induct ts) auto
  finally show ?thesis unfolding weight.simps by (simp add: n_def)
qed

lemma sublisteq_weight:
  assumes "sublisteq ss ts"
  shows "(\<Sum>s\<leftarrow>ss. weight s) \<le> (\<Sum>t\<leftarrow>ts. weight t)"
  using assms by (induct) simp+

text {*Without subterm coefficient functions, @{term S} is a simplification-order.*}
lemma simplification_order_on_UNIV:
  assumes "wpo_on pr_strict\<inverse>\<inverse> UNIV"
  shows "simplification_order_on (\<lambda>x y. S y x) UNIV"
  (is "simplification_order_on ?kbo UNIV")
proof (unfold simplification_order_on_def, intro conjI)
  show "rewrite_order ?kbo"
    unfolding rewrite_order_def rewrite_relation_def po_on_def irreflp_on_def transp_on_def
    by (auto intro: S_subst S_ctxt S_trans simp: S_irrefl simp del: kbo.simps)
next
  let ?P = "\<lambda>x y. pr_strict y x \<and> w x \<le> w y"
  have "range w \<subseteq> UNIV" by auto
  from wpo_on_map [OF assms wpo_on_UNIV_nat this]
    have "wpo_on ?P UNIV" by simp
  moreover
  have "\<forall>s t. emb UNIV ?P s t \<longrightarrow> S t s"
  proof (intro allI impI)
    fix s t :: "('f, 'v) term"
    assume "emb UNIV ?P s t"
    then show "S t s"
    proof (induct)
      case (term_hemb_base f n ts t)
      with S_subterm_one show ?case by blast
    next
      case (term_hemb_sublisteq f n g m ss ts)
      note IH = this
      let ?s = "Fun f ss" and ?t = "Fun g ts"
      from IH
        have strict: "pr_strict (g, length ts) (f, length ss)"
        and w: "w (f, length ss) \<le> w (g, length ts)"
        by auto
      from IH have "?s \<in> gterms UNIV" and "?t \<in> gterms UNIV" by auto
      then have "ground ?s" and "ground ?t" by (blast dest: gtermsD)+
      with ground_imp_vars_term_ms_SCF_empty
        have "vars_term_ms (SCF ?s) \<le> vars_term_ms (SCF ?t)"
        by (metis order_refl)
      moreover from sublisteq_weight [OF `sublisteq ss ts`]
        have "weight ?s \<le> weight ?t" using w by simp
      ultimately show ?case
        using strict by simp
    next
      case (term_hemb_trans s t u)
      then show ?case by (blast intro: S_trans)
    next
      case (term_hemb_ctxt s t f n ss1 ss2)
      with S_mono_one show ?case by blast
    qed
  qed
  ultimately
  show "\<exists>P. wpo_on P UNIV \<and> (\<forall>s t. emb UNIV P s t \<longrightarrow> S t s)" by blast
qed

text {*Alternative termination proof for the case that @{term pr_strict} is a
well-partial-order.*}
lemma SN_S:
  assumes "wpo_on pr_strict\<inverse>\<inverse> UNIV"
  shows "SN {(x :: ('f, 'v) term, y). S x y}"
proof -
  fix c :: 'f
  let ?P = "\<lambda>(x :: ('f, 'v) term) y. S y x"
  have "(c, 0) \<in> UNIV" by auto
  from simplification_order_on_imp_wfp_on_gterms [OF simplification_order_on_UNIV]
    have "wfp_on ?P (gterms UNIV)" using assms by blast
  moreover have "\<And>s t \<sigma>. ?P s t \<Longrightarrow> ?P (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
    using S_subst by best
  ultimately
  have "wfp_on ?P (terms UNIV)"
    using wfp_on_gterms_with_const_imp_wfp_on [of c UNIV ?P] by blast
  then have "wfp_on ?P UNIV" by simp
  then show ?thesis unfolding wfp_on_SN_conv .
qed

declare weight_simp' [simp del]
end

section {* A direct strong normalization proof *}

lemma kbo_strongly_normalizing: fixes s :: "('f,'v)term" 
  shows "SN_on {(s,t). S s t} {s}"
proof -
  let ?SN = "\<lambda> t :: ('f,'v)term. SN_on {(s,t). S s t} {t}"
  let ?m1 = "\<lambda> (f,ss). weight (Fun f ss)"
  let ?m2 = "\<lambda> (f,ss). (f,length ss)"
  let ?rel' = "lex_two (\<lambda> fss gts. ?m1 fss > ?m1 gts) (\<lambda> fss gts. ?m1 fss \<ge> ?m1 gts) (\<lambda> fss gts. pr_strict (?m2 fss) (?m2 gts))"
  let ?rel = "\<lambda> fss gts. ?rel' (fss,fss) (gts,gts)"
  have SN_rel': "SN {(fss,gts) . ?rel' fss gts}"
    by (rule lex_two, insert SN_inv_image[OF pr_SN, of ?m2]  SN_inv_image[OF SN_nat_gt, of ?m1],
      auto simp: inv_image_def)
  from SN_inv_image[OF this, of "\<lambda> x. (x,x)"] have SN_rel: "SN {(fss,gts). ?rel fss gts}"
    unfolding inv_image_def by auto
  note conv = SN_on_all_reducts_SN_on_conv
  show "?SN s"
  proof (induct s)
    case (Var x)
    show ?case unfolding conv[of _ "Var x"] using not_S_Var[of x] by auto
  next
    case (Fun f ss)
    hence subset: "set ss \<subseteq> {s. ?SN s}" by blast
    let ?P = "\<lambda> (f,ss). set ss \<subseteq> {s. ?SN s} \<longrightarrow> ?SN (Fun f ss)"
    {
      fix fss
      have "?P fss"      
      proof (induct fss rule: SN_induct[OF SN_rel])
        case (1 fss)
        obtain f ss where fss: "fss = (f,ss)" by force
        {
          fix g ts
          assume "?m1 (f,ss) > ?m1 (g,ts) \<or> ?m1 (f,ss) \<ge> ?m1 (g,ts) \<and> pr_strict (?m2 (f,ss)) (?m2 (g,ts))"
            and "set ts \<subseteq> {s. ?SN s}"
          hence "?SN (Fun g ts)"
            using 1[rule_format, of "(g,ts)", unfolded fss split] by auto
        } note IH = this[unfolded split]
        show ?case unfolding fss split
        proof
          assume SN_s: "set ss \<subseteq> {s. ?SN s}"
          let ?f = "(f,length ss)"
          let ?s = "Fun f ss"
          let ?SNt = "\<lambda> g ts. ?SN (Fun g ts)"
          let ?sym = "\<lambda> g ts. (g,length ts)"
          let ?lex = "lex_ext kbo (weight ?s)"
          let ?lexu = "lex_ext_unbounded kbo"
          let ?lex_SN = "{(ys,xs). (\<forall> y \<in> set ys. ?SN y) \<and> fst (?lex ys xs)}"
          from lex_ext_SN[of kbo "weight ?s", OF S_imp_NS NS_S_compat]
          have SN: "SN ?lex_SN" .
          {
            fix g and ts :: "('f,'v)term list"
            assume "pr_weak ?f (?sym g ts) \<and> weight (Fun g ts) \<le> weight ?s \<and> set ts \<subseteq> {s. ?SN s}"
            hence "?SNt g ts"
            proof (induct ts arbitrary: g rule: SN_induct[OF SN])
              case (1 ts g)
              note inner_IH = 1(1)
              let ?g = "(g,length ts)"
              let ?t = "Fun g ts"
              from 1(2) have fg: "pr_weak ?f ?g" and w: "weight ?t \<le> weight ?s"  and SN: "set ts \<subseteq> {s. ?SN s}" by auto
              show "?SNt g ts" unfolding conv[of _ ?t]
              proof (intro allI impI)
                fix u
                assume "(?t,u) \<in> {(s,t). S s t}"
                hence tu: "S ?t u" by auto
                thus "?SN u"
                proof (induct u)
                  case (Var x)
                  thus ?case using not_S_Var[of x] unfolding conv[of _ "Var x"] by auto
                next
                  case (Fun h us) 
                  let ?h = "(h,length us)"
                  let ?u = "Fun h us"
                  note tu = Fun(2)
                  {
                    fix u 
                    assume u: "u \<in> set us"
                    hence "?u \<rhd> u" by auto
                    from S_trans[OF tu kbo_supt[OF this]] have "S ?t u" by auto
                    from Fun(1)[OF u this] have "?SN u" .
                  } hence SNu: "set us \<subseteq> {s . ?SN s}" by blast
                  note IH = IH[OF _ this]
                  from tu have wut: "weight ?u \<le> weight ?t" by (simp split: if_splits)
                  show ?case 
                  proof (cases "?m1 (f,ss) > ?m1 (h,us) \<or> ?m1 (f,ss) \<ge> ?m1 (h,us) \<and> pr_strict (?m2 (f,ss)) (?m2 (h,us))")
                    case True
                    from IH[OF True[unfolded split]] show ?thesis by simp
                  next
                    case False
                    with wut w have wut: "weight ?t = weight ?u" "weight ?s = weight ?u" by auto
                    note False = False[unfolded split wut]
                    note tu = tu[unfolded kbo.simps wut, unfolded Fun term.simps, simplified]
                    from tu have gh: "pr_weak ?g ?h" unfolding pr_strict by (auto split: if_splits)
                    from pr_weak_trans[OF fg gh] have fh: "pr_weak ?f ?h" .
                    from False wut fh have "\<not> pr_strict ?f ?h" unfolding pr_strict by auto
                    with fh have hf: "pr_weak ?h ?f" unfolding pr_strict by auto
                    from pr_weak_trans[OF hf fg] have hg: "pr_weak ?h ?g" .
                    from hg have gh2: "\<not> pr_strict ?g ?h" unfolding pr_strict by auto
                    from tu gh2 have lex: "fst (?lexu ts us)" by (auto split: if_splits)
                    from fh wut SNu have "pr_weak ?f ?h \<and> weight ?u \<le> weight ?s \<and> set us \<subseteq> {s. ?SN s}" 
                      by auto
                    note inner_IH = inner_IH[OF _ this]
                    show ?thesis 
                    proof (rule inner_IH, rule, unfold split, intro conjI ballI)
                      have "fst (?lexu ts us)" by (rule lex)
                      moreover have "length us \<le> weight ?s" 
                      proof -
                        have "length us \<le> listsum (map weight us)" 
                        proof (induct us)
                          case (Cons u us)
                          from Cons have "length (u # us) \<le> Suc (listsum (map weight us))" by auto
                          also have "... \<le> listsum (map weight (u # us))" using weight_gt_0[of u]
                            by auto
                          finally show ?case .
                        qed simp
                        also have "\<dots> \<le> listsum (map weight (scf_list (scf (h,length us)) us))"
                          by (rule listsum_scf_list[OF scf])
                        also have "... \<le> weight ?s" using wut by simp
                        finally show ?thesis .
                      qed                          
                      ultimately show "fst (?lex ts us)" unfolding lex_ext_def Let_def by auto
                    next
                      fix t
                      assume t: "t \<in> set ts" 
                      with SN show "?SN t" by blast
                    qed
                  qed
                qed
              qed
            qed
          }
          from this[of f ss] SN_s  show "?SN ?s" by auto
        qed
      qed
    }
    from this[of "(f,ss)", unfolded split]
    show ?case using Fun by blast
  qed
qed

declare kbo.simps[simp del]

section {* use KBO as ce-compatible reduction pair / triple *}

abbreviation kbo_S where "kbo_S \<equiv> {(s,t). S s t}"
abbreviation kbo_NS where "kbo_NS \<equiv> {(s,t). NS s t}"

lemma kbo_ce_compat: "ce_trs x \<subseteq> kbo_S \<inter> kbo_NS"
proof (cases x)
  case (fields a b c)
  {
    fix s t
    assume "(s,t) \<in> ce_trs x"
    hence "s \<rhd> t" unfolding fields ce_trs.simps by auto
    from kbo_supt[OF this] have "(s,t) \<in> kbo_S" by simp
    with S_imp_NS[of s t] have "(s,t) \<in> kbo_S \<inter> kbo_NS" by auto
  }
  thus ?thesis by auto
qed

lemma S_ce_compat: "ce_trs x \<subseteq> kbo_S" using kbo_ce_compat by blast
lemma NS_ce_compat: "ce_trs x \<subseteq> kbo_NS" using kbo_ce_compat by blast

lemma kbo_redpair: "mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af"
proof (unfold_locales)
  show "SN kbo_S" using kbo_strongly_normalizing unfolding SN_defs by blast
  show "subst_closed kbo_S" using S_subst by auto
  show "subst_closed kbo_NS" using NS_subst by auto
  show "ctxt_closed kbo_NS" 
    by (rule, insert NS_ctxt, blast)
  show "kbo_S O kbo_NS \<subseteq> kbo_S" using S_NS_compat by auto
  show "kbo_NS O kbo_S \<subseteq> kbo_S" using NS_S_compat by auto
  show "trans kbo_S" using S_trans unfolding trans_def by auto
  show "trans kbo_NS" using NS_trans unfolding trans_def by auto
  show "refl kbo_NS" using NS_refl unfolding refl_on_def by auto
  show "af_compatible full_af kbo_NS" by (rule full_af)
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d . ce_trs (c,d,m) \<subseteq> kbo_NS"
    using NS_ce_compat by blast
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d . ce_trs (c,d,m) \<subseteq> kbo_S"
    using S_ce_compat by blast
qed

lemma kbo_S_mono: "ctxt_closed kbo_S"
  by (rule, insert S_ctxt, blast)
end

end

